@@ -1,7 +1,7 @@
 /**
  * Copyright (c) 2025, Fabian Groffen. All rights reserved.
  *
- * See LICENSE for the license.
+ * See LICENSE for the licence.
  */
 
 #include <stdio.h>
@@ -15,6 +15,36 @@
 
 #include "util.h"
 
+typedef struct rdflist {
+    ldns_rdf       *d;
+    ldns_rr_type   *r;
+    struct rdflist *next;
+} rdf_list;
+
+typedef struct hostopts {
+    uint8_t        verbose:1;
+    uint8_t        doafxr:1;
+    uint8_t        doip6int:1;
+    uint8_t        recurse:1;
+    uint8_t        failfast:1;
+    uint8_t        checkrev:1;
+    uint8_t        implicit:1;
+    uint8_t        colour:1;
+    uint8_t        ipmode:2;
+#define HOST_IP_UNSET 0
+#define HOST_IP_4     1
+#define HOST_IP_6     2
+    uint8_t        tcpudp:2;
+#define HOST_MODE_UNSET 0
+#define HOST_MODE_TCP   1
+#define HOST_MODE_UDP   2
+    uint8_t        ndots;
+    uint8_t        nretries;
+    uint8_t        timeout;
+    uint16_t       port;
+    ldns_rr_class  qclass;
+} host_opts;
+
 static int
 host_print_record
 (
@@ -44,8 +74,10 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s has %saddress %s\n",
-                    own, tpe == LDNS_RR_TYPE_AAAA ? "IPv6 " : "", dest);
+            fprintf(stdout, "%s%s%s has %saddress %s%s%s\n",
+                    FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                    tpe == LDNS_RR_TYPE_AAAA ? "IPv6 " : "",
+                    FGCOLOUR_BRGREEN, dest, FGCOLOUR_NORM);
             ret = EXIT_SUCCESS;
 
             free(dest);
@@ -55,7 +87,9 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s is an alias for %s\n", own, dest);
+            fprintf(stdout, "%s%s%s is an alias for %s%s%s\n",
+                    FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                    FGCOLOUR_BRBLUE, dest, FGCOLOUR_NORM);
             ret = EXIT_SUCCESS;
 
             free(dest);
@@ -89,8 +123,10 @@ host_print_record
             }
             if (trg != NULL)
             {
-                fprintf(stdout, "%s mail is handled by %u %s\n",
-                        own, prio, trg);
+                fprintf(stdout, "%s%s%s mail is handled by %s%u %s%s%s\n",
+                        FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                        FGCOLOUR_BRYELLOW, prio,
+                        FGCOLOUR_BRCYAN, trg, FGCOLOUR_NORM);
                 ret = EXIT_SUCCESS;
             }
 
@@ -101,7 +137,9 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s domain name pointer %s\n", own, dest);
+            fprintf(stdout, "%s%s%s domain name pointer %s%s%s\n",
+                    FGCOLOUR_GREEN, own, FGCOLOUR_NORM,
+                    FGCOLOUR_BRBLUE, dest, FGCOLOUR_NORM);
             ret = EXIT_SUCCESS;
 
             free(dest);
@@ -111,7 +149,9 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s descriptive text %s\n", own, dest);
+            fprintf(stdout, "%s%s%s descriptive text %s%s%s\n",
+                    FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                    FGCOLOUR_BRMAGENTA, dest, FGCOLOUR_NORM);
             ret = EXIT_SUCCESS;
 
             free(dest);
@@ -121,7 +161,9 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s name server %s\n", own, dest);
+            fprintf(stdout, "%s%s%s name server %s%s%s\n",
+                    FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                    FGCOLOUR_BRGREEN, dest, FGCOLOUR_NORM);
             ret = EXIT_SUCCESS;
 
             free(dest);
@@ -131,7 +173,9 @@ host_print_record
             ldns_rdf *targ = ldns_rr_rdf(rr, 0);
             char     *dest = ldns_rdf2str(targ);
 
-            fprintf(stdout, "%s has SOA record %s", own, dest);
+            fprintf(stdout, "%s%s%s has SOA record %s%s%s",
+                    FGCOLOUR_BLUE, own, FGCOLOUR_NORM,
+                    FGCOLOUR_BRMAGENTA, dest, FGCOLOUR_NORM);
             free(dest);
             for (j = 1; j < ldns_rr_rd_count(rr); j++)
             {
@@ -145,21 +189,28 @@ host_print_record
                     case LDNS_RDF_TYPE_DNAME:
                     case LDNS_RDF_TYPE_STR:
                         dest = ldns_rdf2str(targ);
-                        fprintf(stdout, " %s", dest);
+                        fprintf(stdout, " %s%s%s",
+                                FGCOLOUR_BRCYAN, dest, FGCOLOUR_NORM);
                         free(dest);
                         break;
                     case LDNS_RDF_TYPE_INT8:
-                        fprintf(stdout, " %u",
-                                (uint32_t)ldns_rdf2native_int8(targ));
+                        fprintf(stdout, " %s%u%s",
+                                FGCOLOUR_BRYELLOW,
+                                (uint32_t)ldns_rdf2native_int8(targ),
+                                FGCOLOUR_NORM);
                         break;
                     case LDNS_RDF_TYPE_INT16:
-                        fprintf(stdout, " %u",
-                                (uint32_t)ldns_rdf2native_int16(targ));
+                        fprintf(stdout, " %s%u%s",
+                                FGCOLOUR_YELLOW,
+                                (uint32_t)ldns_rdf2native_int16(targ),
+                                FGCOLOUR_NORM);
                         break;
                     case LDNS_RDF_TYPE_INT32:
                     case LDNS_RDF_TYPE_PERIOD:
-                        fprintf(stdout, " %u",
-                                ldns_rdf2native_int32(targ));
+                        fprintf(stdout, " %s%u%s",
+                                FGCOLOUR_GREEN,
+                                ldns_rdf2native_int32(targ),
+                                FGCOLOUR_NORM);
                         break;
                     default:
                         /* ignore this RD */
@@ -181,11 +232,9 @@ static int
 host_func
 (
     ldns_resolver *res,
+    host_opts     *opts,
     ldns_rdf      *domain,
-    ldns_rr_type  *types,
-    ldns_rr_class  qclass,
-    bool           verbose,
-    bool           implicit
+    ldns_rr_type  *types
 )
 {
     ldns_pkt      *pkt;
@@ -196,17 +245,37 @@ host_func
 
     while ((uint32_t)(tpe = *types++) != 0)
     {
-        pkt = ldns_resolver_search(res, domain, tpe, qclass, LDNS_RD);
+        ldns_status resret =
+            ldns_resolver_search_status(&pkt, res, domain,
+                                        tpe, opts->qclass, LDNS_RD);
 
-        if (pkt == NULL)
+        switch (resret)
         {
-            fprintf(stderr,
-                    ";; connection timed out; no servers could be reached\n");
+            case LDNS_STATUS_NETWORK_ERR:
+            case LDNS_STATUS_ADDRESS_ERR:
+            case LDNS_STATUS_SOCKET_ERROR:
+                fprintf(stderr,
+                        "%s;; %s%s%s; no servers could be reached%s\n",
+                        FGCOLOUR_RED, FGCOLOUR_BRRED,
+                        ldns_get_errorstr_by_id(resret),
+                        FGCOLOUR_RED, FGCOLOUR_NORM);
 
-            return EXIT_FAILURE;
+                return EXIT_FAILURE;
+            case LDNS_STATUS_OK:
+                /* all good, let's do something */
+                /* note: pkt can be NULL in this case :( */
+                break;
+            default:
+                fprintf(stderr, "%s;; %s%s%s\n",
+                        FGCOLOUR_RED, FGCOLOUR_BRRED,
+                        ldns_get_errorstr_by_id(resret),
+                        FGCOLOUR_NORM);
+                return EXIT_FAILURE;
         }
-        else if (ldns_pkt_reply_type(pkt) != LDNS_PACKET_ANSWER &&
-                 ldns_pkt_reply_type(pkt) != LDNS_PACKET_NODATA)
+
+        if (pkt == NULL ||
+            (ldns_pkt_reply_type(pkt) != LDNS_PACKET_ANSWER &&
+             ldns_pkt_reply_type(pkt) != LDNS_PACKET_NODATA))
         {
             char           *dom;
             char           *rcd;
@@ -221,8 +290,9 @@ host_func
             dom = ldns_rdf2str(domain);
             rcd = ldns_pkt_rcode2str(rc);
 
-            fprintf(stderr, "Host %s not found: %u(%s)\n",
-                    dom, (uint32_t)rc, rcd);
+            fprintf(stderr, "%sHost %s%s%s not found: %s%u(%s)%s\n",
+                    FGCOLOUR_RED, FGCOLOUR_BRRED, dom, FGCOLOUR_RED,
+                    FGCOLOUR_BRBLUE, (uint32_t)rc, rcd, FGCOLOUR_NORM);
 
             free(dom);
             free(rcd);
@@ -232,7 +302,7 @@ host_func
             return EXIT_FAILURE;
         }
 
-        if (verbose)
+        if (opts->verbose)
         {
             char *digd = ldns_pkt2str(pkt);
             fprintf(stdout, "%s\n", digd);
@@ -246,7 +316,7 @@ host_func
                 ldns_rr_list_sort(ans);
             if (host_print_record(ans) == EXIT_SUCCESS)
                 found = true;
-            ldns_rr_list_deep_free(ans);
+            ldns_rr_list_free(ans);
         }
 
         ldns_pkt_free(pkt);
@@ -254,7 +324,7 @@ host_func
         lasttpe = tpe;
     }
 
-    if (!implicit &&
+    if (!opts->implicit &&
         !found &&
         lasttpe != (ldns_rr_type)0)
     {
@@ -268,7 +338,9 @@ host_func
         /* take last type to report on */
         rr  = ldns_rr_type2str(lasttpe);
 
-        fprintf(stderr, "%s has no %s record\n", dom, rr);
+        fprintf(stderr, "%s%s%s has no %s%s%s record%s\n",
+                FGCOLOUR_BRRED, dom, FGCOLOUR_RED,
+                FGCOLOUR_BRBLUE, rr, FGCOLOUR_RED, FGCOLOUR_NORM);
 
         free(dom);
         free(rr);
@@ -279,19 +351,214 @@ host_func
     return EXIT_SUCCESS;
 }
 
+static int
+host_checkptr
+(
+    ldns_resolver *res,
+    host_opts     *opts,
+    rdf_list      *domains,
+    rdf_list      *seen
+)
+{
+    ldns_pkt      *pkt;
+    ldns_rr_list  *ans;
+    ldns_rr_type  *qtypes = domains->r;
+    ldns_rr_type  *types;
+    ldns_rr_type   tpe;
+    ldns_rdf      *revq;
+    rdf_list      *seent;
+    rdf_list      *todot = domains;
+    rdf_list      *nextd;
+    int            i;
+
+    for (seent = seen; seent->next != NULL; seent = seent->next)
+        ;
+
+    /* switch to default if input is an address */
+    if (qtypes[0] == LDNS_RR_TYPE_PTR)
+    {
+        qtypes = calloc(3, sizeof(qtypes[0]));
+        qtypes[0] = LDNS_RR_TYPE_A;
+        qtypes[1] = LDNS_RR_TYPE_AAAA;
+    }
+
+    while (domains != NULL)
+    {
+        bool addseen = true;
+
+        types = domains->r;
+
+        while ((uint32_t)(tpe = *types++) != 0)
+        {
+            ldns_status resret =
+                ldns_resolver_search_status(&pkt, res, domains->d,
+                                            tpe, opts->qclass, LDNS_RD);
+
+            switch (resret)
+            {
+                case LDNS_STATUS_NETWORK_ERR:
+                case LDNS_STATUS_ADDRESS_ERR:
+                case LDNS_STATUS_SOCKET_ERROR:
+                    fprintf(stderr,
+                            "%s;; %s%s%s; no servers could be reached%s\n",
+                            FGCOLOUR_RED, FGCOLOUR_BRRED,
+                            ldns_get_errorstr_by_id(resret),
+                            FGCOLOUR_RED, FGCOLOUR_NORM);
+
+                    return EXIT_FAILURE;
+                case LDNS_STATUS_OK:
+                    /* all good, let's do something */
+                    /* note: pkt can be NULL in this case :( */
+                    break;
+                default:
+                    fprintf(stderr, "%s;; %s%s%s\n",
+                            FGCOLOUR_RED, FGCOLOUR_BRRED,
+                            ldns_get_errorstr_by_id(resret),
+                            FGCOLOUR_NORM);
+                    return EXIT_FAILURE;
+            }
+
+            ans = NULL;
+            if (pkt == NULL ||
+                (ldns_pkt_reply_type(pkt) != LDNS_PACKET_ANSWER &&
+                 ldns_pkt_reply_type(pkt) != LDNS_PACKET_NODATA))
+            {
+                char           *dom;
+                char           *rcd;
+                ldns_pkt_rcode  rc;
+
+                rc = LDNS_RCODE_NXDOMAIN;
+                if (pkt != NULL)
+                    rc = ldns_pkt_get_rcode(pkt);
+
+                if (ldns_dname_absolute(domains->d))
+                    ldns_rdf_set_size(domains->d,
+                                      ldns_rdf_size(domains->d) - 1);
+                dom = ldns_rdf2str(domains->d);
+                rcd = ldns_pkt_rcode2str(rc);
+
+                fprintf(stderr, "%sHost %s%s%s not found: %s%u(%s)%s\n",
+                        FGCOLOUR_RED, FGCOLOUR_BRRED, dom, FGCOLOUR_RED,
+                        FGCOLOUR_BRBLUE, (uint32_t)rc, rcd, FGCOLOUR_NORM);
+
+                free(dom);
+                free(rcd);
+            }
+            else
+            {
+                ans = ldns_pkt_rr_list_by_type(pkt, tpe, LDNS_SECTION_ANSWER);
+            }
+
+            if (ans != NULL)
+            {
+                if (addseen)
+                {
+                    ldns_rr *rr = ldns_rr_list_rr(ans, 0);
+
+                    /* append this domain to the "seen" list, we delay this
+                     * until we have a resolve, because transformations may
+                     * happen due to non-absolute input, and added search
+                     * domains, by delaying until here, we can properly
+                     * ignore re-resolving the same domain names lateron */
+                    seent    = seent->next = calloc(1, sizeof(*seent));
+                    seent->d = ldns_rdf_clone(ldns_rr_owner(rr));
+
+                    addseen  = false;
+                }
+
+                if (host_print_record(ans) == EXIT_SUCCESS)
+                {
+                    for (i = 0; i < ldns_rr_list_rr_count(ans); i++)
+                    {
+                        ldns_rr       *rr   = ldns_rr_list_rr(ans, i);
+                        ldns_rr_type   tpe  = ldns_rr_get_type(rr);
+                        ldns_rdf      *rrdf = ldns_rr_rdf(rr, 0);
+                        ldns_rr_type  *ntpe;
+
+                        switch (tpe)
+                        {
+                            case LDNS_RR_TYPE_A:
+                            case LDNS_RR_TYPE_AAAA:
+                                revq = util_addr2dname(rrdf, false);
+                                ntpe = calloc(2, sizeof(ntpe[0]));
+                                ntpe[0] = LDNS_RR_TYPE_PTR;
+                                break;
+                            case LDNS_RR_TYPE_PTR:
+                                revq = rrdf;
+                                ntpe = qtypes;
+                                break;
+                            default:
+                                exit(EXIT_FAILURE);
+                        }
+
+                        for (nextd = seen->next;
+                             nextd != NULL;
+                             nextd = nextd->next)
+                        {
+                            if (ldns_rdf_compare(nextd->d, revq) == 0)
+                            {
+                                /* already seen, skip */
+                                ldns_rdf_deep_free(revq);
+                                if (ntpe != qtypes)
+                                    free(ntpe);
+                                revq = NULL;
+                                break;
+                            }
+                        }
+                        for (nextd = domains;
+                             nextd != NULL;
+                             nextd = nextd->next)
+                        {
+                            if (ldns_rdf_compare(nextd->d, revq) == 0)
+                            {
+                                /* already scheduled, skip */
+                                ldns_rdf_deep_free(revq);
+                                if (ntpe != qtypes)
+                                    free(ntpe);
+                                revq = NULL;
+                                break;
+                            }
+                        }
+
+                        if (revq != NULL)
+                        {
+                            todot = todot->next = calloc(1, sizeof(*todot));
+                            todot->d = revq;
+                            todot->r = ntpe;
+                        }
+                    }
+                }
+                ldns_rr_list_free(ans);
+            }
+
+            ldns_pkt_free(pkt);
+        }
+
+        if (domains->r != qtypes)
+            free(domains->r);
+        free(domains->d);
+        nextd = domains->next;
+        free(domains);
+        domains = nextd;
+    }
+
+    free(qtypes);
+
+    return EXIT_SUCCESS;
+}
+
 static int
 host_afxr
 (
     ldns_resolver *res,
     ldns_rdf      *domain,
     ldns_rr_type  *types,
-    ldns_rr_class  qclass,
-    bool           verbose
+    host_opts     *opts
 )
 {
     ldns_rr_list *ans;
 
-    if (ldns_axfr_start(res, domain, qclass) != LDNS_STATUS_OK)
+    if (ldns_axfr_start(res, domain, opts->qclass) != LDNS_STATUS_OK)
         return EXIT_FAILURE;
 
     ans = ldns_rr_list_new();
@@ -318,8 +585,9 @@ host_afxr
             dom = ldns_rdf2str(domain);
             rcd = ldns_pkt_rcode2str(rc);
 
-            fprintf(stderr, "Host %s not found: %u(%s)\n",
-                    dom, (uint32_t)rc, rcd);
+            fprintf(stderr, "%sHost %s%s%s not found: %s%u(%s)%s\n",
+                    FGCOLOUR_RED, FGCOLOUR_BRRED, dom, FGCOLOUR_RED,
+                    FGCOLOUR_BRBLUE, (uint32_t)rc, rcd, FGCOLOUR_NORM);
 
             free(dom);
             free(rcd);
@@ -340,7 +608,7 @@ host_afxr
         }
     }
 
-    if (verbose)
+    if (opts->verbose)
     {
         size_t i;
         for (i = 0; i < ldns_rr_list_rr_count(ans); i++)
@@ -389,6 +657,10 @@ do_usage
 "    -W specifies how long to wait for a reply\n"
 "    -4 use IPv4 query transport only\n"
 "    -6 use IPv6 query transport only\n"
+"  ldns-tools extensions\n"
+"    -h this help screen\n"
+"    -C use colours (defaults to true on ttys, respects NO_COLOR envvar)\n"
+"    -P resolve resulting address and pointer answers (A/AAAA/PTR consistency)\n"
 );
 }
 
@@ -399,30 +671,35 @@ main(int argc, char **argv)
     ldns_rdf      *domain;
     ldns_rr_type  *qtypes   = NULL;
     ldns_status    s;
-    ldns_rr_class  qclass   = LDNS_RR_CLASS_IN;
-    int            opt;
     char          *typestr  = NULL;
     char          *classstr = NULL;
-    uint8_t        ndots    = 1;
-    uint8_t        nretries = 1;
-    uint8_t        ipmode   = 0;  /* no preference */
-    uint8_t        timeout  = 1;  /* second */
-    uint8_t        tcpudp   = 0;  /* no preference */
-    uint16_t       port     = 0;
-    bool           verbose  = false;
-    bool           doafxr   = false;
-    bool           doip6int = false;  /* stick with ip6.arpa */
-    bool           recurse  = true;
-    bool           failfast = false;  /* try all nameservers */
+    char          *nocolour = getenv("NO_COLOR");
+    int            opt;
+    host_opts      opts = {
+        .ndots    = 1,
+        .nretries = 1,
+        .timeout  = 1,  /* second */
+        .recurse  = true,
+        .colour   = true,
+        .qclass   = LDNS_RR_CLASS_IN
+    };
 
-    while ((opt = getopt(argc, argv, "46aAc:dilN:p:srR:t:TUvVwW:")) != -1)
+    /* https://no-color.org/ */
+    if (nocolour != NULL &&
+        *nocolour != '\0')
+        opts.colour = false;
+
+    if (!isatty(fileno(stdout)))
+        opts.colour = false;
+
+    while ((opt = getopt(argc, argv, "46aAc:CdhilN:p:PsrR:t:TUvVwW:")) != -1)
     {
         switch (opt) {
             case '4':
-                ipmode = 1;
+                opts.ipmode = HOST_IP_4;
                 break;
             case '6':
-                ipmode = 2;
+                opts.ipmode = HOST_IP_6;
                 break;
             case 'a':
             case 'A':  /* should be a - [RRSIG, NSEC, NSEC3] */
@@ -430,16 +707,22 @@ main(int argc, char **argv)
                 /* fall through */
             case 'd':
             case 'v':
-                verbose = true;
+                opts.verbose = true;
                 break;
             case 'c':
                 classstr = optarg;
                 break;
+            case 'C':
+                opts.colour = true;
+                break;
+            case 'h':
+                do_usage(stdout);
+                exit(EXIT_SUCCESS);
             case 'i':
-                doip6int = true;
+                opts.doip6int = true;
                 break;
             case 'l':
-                doafxr = true;
+                opts.doafxr = true;
                 break;
             case 'N':
                 {
@@ -464,7 +747,7 @@ main(int argc, char **argv)
                         exit(EXIT_FAILURE);
                     }
 
-                    ndots = (uint8_t)val;
+                    opts.ndots = (uint8_t)val;
                 }
                 break;
             case 'p':
@@ -490,11 +773,14 @@ main(int argc, char **argv)
                         exit(EXIT_FAILURE);
                     }
 
-                    port = (uint16_t)val;
+                    opts.port = (uint16_t)val;
                 }
                 break;
+            case 'P':
+                opts.checkrev = true;
+                break;
             case 'r':
-                recurse = false;
+                opts.recurse = false;
                 break;
             case 'R':
                 {
@@ -519,29 +805,29 @@ main(int argc, char **argv)
                     }
 
                     if (val <= 0)
-                        nretries = 1;
+                        opts.nretries = 1;
                     else
-                        nretries = (uint8_t)val;
+                        opts.nretries = (uint8_t)val;
                 }
                 break;
             case 's':
-                failfast = true;
+                opts.failfast = true;
                 break;
             case 't':
                 typestr = optarg;
                 break;
             case 'U':
-                tcpudp = 2;
+                opts.tcpudp = HOST_MODE_UDP;
                 break;
             case 'T':
-                tcpudp = 1;
+                opts.tcpudp = HOST_MODE_TCP;
                 break;
             case 'V':
                 fprintf(stdout, "host %s from ldns-tools (using ldns %s)\n",
                         LDNS_TOOLS_VERSION, ldns_version());
                 exit(EXIT_SUCCESS);
             case 'w':
-                timeout = 0;  /* forever */
+                opts.timeout = 0;  /* forever */
                 break;
             case 'W':
                 {
@@ -566,9 +852,9 @@ main(int argc, char **argv)
                     }
 
                     if (val <= 0)
-                        timeout = 1;
+                        opts.timeout = 1;
                     else
-                        timeout = (uint8_t)val;
+                        opts.timeout = (uint8_t)val;
                 }
                 break;
             default:
@@ -577,6 +863,8 @@ main(int argc, char **argv)
         }
     }
 
+    util_enable_colours(opts.colour);
+
     if (optind >= argc)
     {
         do_usage(stdout);
@@ -614,16 +902,31 @@ main(int argc, char **argv)
                 qtypes[0] = rrtype;
                 break;
         }
+
+        if (opts.checkrev)
+        {
+            switch (rrtype)
+            {
+                case LDNS_RR_TYPE_A:
+                case LDNS_RR_TYPE_AAAA:
+                case LDNS_RR_TYPE_PTR:
+                    /* ok, accept this */
+                    break;
+                default:
+                    fprintf(stderr, "host: -P requires A, AAAA or PTR type\n");
+                    exit(EXIT_FAILURE);
+            }
+        }
     }
 
     if (classstr != NULL)
     {
         if (strcasecmp(classstr, "in") == 0)
-            qclass = LDNS_RR_CLASS_IN;
+            opts.qclass = LDNS_RR_CLASS_IN;
         else if (strcasecmp(classstr, "ch") == 0)
-            qclass = LDNS_RR_CLASS_CH;
+            opts.qclass = LDNS_RR_CLASS_CH;
         else if (strcasecmp(classstr, "hs") == 0)
-            qclass = LDNS_RR_CLASS_HS;
+            opts.qclass = LDNS_RR_CLASS_HS;
         else
         {
             fprintf(stderr, "host: invalid class: %s\n", classstr);
@@ -637,7 +940,7 @@ main(int argc, char **argv)
     /* take second argument as DNS server to use */
     if (argc - optind > 1)
     {
-        ldns_rdf **dns = util_addr_frm_str(res, argv[optind + 1], ndots);
+        ldns_rdf **dns = util_addr_frm_str(res, argv[optind + 1], opts.ndots);
 
         if (dns == NULL)
         {
@@ -678,48 +981,49 @@ main(int argc, char **argv)
         exit(EXIT_FAILURE);
     }
 
-    if (port > 0)
-        ldns_resolver_set_port(res, port);
-    ldns_resolver_set_recursive(res, recurse);
-    ldns_resolver_set_retry(res, nretries);
-    ldns_resolver_set_ip6(res, ipmode);
+    if (opts.port > 0)
+        ldns_resolver_set_port(res, opts.port);
+    ldns_resolver_set_recursive(res, opts.recurse);
+    ldns_resolver_set_retry(res, opts.nretries);
+    ldns_resolver_set_ip6(res, opts.ipmode);
     {
         struct timeval tvtimeout = {
-            .tv_sec = timeout == 0 ? INT_MAX : timeout
+            .tv_sec = opts.timeout == 0 ? INT_MAX : opts.timeout
         };
         ldns_resolver_set_timeout(res, tvtimeout);
-        ldns_resolver_set_retrans(res, timeout);
+        ldns_resolver_set_retrans(res, opts.timeout);
     }
-    if (tcpudp > 0)
-        ldns_resolver_set_usevc(res, tcpudp == 1 ? true : false);
-    ldns_resolver_set_fail(res, failfast);
+    if (opts.tcpudp > 0)
+        ldns_resolver_set_usevc(res,
+                                opts.tcpudp == HOST_MODE_TCP ? true : false);
+    ldns_resolver_set_fail(res, opts.failfast);
 
     s = ldns_str2rdf_a(&domain, argv[optind]);
     if (s != LDNS_STATUS_OK)
         s = ldns_str2rdf_aaaa(&domain, argv[optind]);
     if (s == LDNS_STATUS_OK)
     {
-        if (!doafxr &&
+        if (!opts.doafxr &&
             qtypes == NULL)
         {
             qtypes = calloc(sizeof(qtypes[0]), 2);
             qtypes[0] = LDNS_RR_TYPE_PTR;
         }
 
-        domain = util_addr2dname(domain, doip6int);
+        domain = util_addr2dname(domain, opts.doip6int);
     }
     else
     {
-        domain = util_dname_frm_str(argv[optind], ndots, doip6int);
+        domain = util_dname_frm_str(argv[optind], opts.ndots, opts.doip6int);
     }
 
-    if (verbose)
+    if (opts.verbose)
     {
         char *dom = ldns_rdf2str(domain);
         fprintf(stdout, "Trying \"%s\"\n", dom);
         free(dom);
     }
-    if (doafxr)
+    if (opts.doafxr)
     {
         if (qtypes == NULL)
         {
@@ -729,26 +1033,57 @@ main(int argc, char **argv)
             qtypes[2] = LDNS_RR_TYPE_AAAA;
             qtypes[3] = LDNS_RR_TYPE_PTR;
         }
-        host_afxr(res, domain, qtypes, qclass, verbose);
+        host_afxr(res, domain, qtypes, &opts);
     }
-    else
+    else if (opts.checkrev)
     {
-        bool implicit = false;
+        rdf_list  seenlist = { .d = NULL, .next = NULL };
+        rdf_list *seenw    = &seenlist;
+        rdf_list *todo;
 
         if (qtypes == NULL)
         {
-            qtypes = calloc(sizeof(qtypes[0]), 5);
-            qtypes[0] = LDNS_RR_TYPE_CNAME;
-            qtypes[1] = LDNS_RR_TYPE_A;
-            qtypes[2] = LDNS_RR_TYPE_AAAA;
-            qtypes[3] = LDNS_RR_TYPE_MX;
-            implicit  = true;
+            qtypes = calloc(sizeof(qtypes[0]), 3);
+            qtypes[0] = LDNS_RR_TYPE_A;
+            qtypes[1] = LDNS_RR_TYPE_AAAA;
         }
-        host_func(res, domain, qtypes, qclass, verbose, implicit);
+
+        todo = calloc(1, sizeof(*todo));
+        todo->d    = domain;
+        todo->r    = qtypes;
+        todo->next = NULL;
+        domain     = NULL;
+        qtypes     = NULL;
+
+        host_checkptr(res, &opts, todo, seenw);
+
+        seenw = seenw->next;
+        while (seenw != NULL)
+        {
+            rdf_list *next = seenw->next;
+            if (seenw->d != domain)
+                ldns_rdf_deep_free(seenw->d);
+            free(seenw);
+            seenw = next;
+        }
+    }
+    else
+    {
+        if (qtypes == NULL)
+        {
+            qtypes        = calloc(sizeof(qtypes[0]), 5);
+            qtypes[0]     = LDNS_RR_TYPE_CNAME;
+            qtypes[1]     = LDNS_RR_TYPE_A;
+            qtypes[2]     = LDNS_RR_TYPE_AAAA;
+            qtypes[3]     = LDNS_RR_TYPE_MX;
+            opts.implicit = true;
+        }
+        host_func(res, &opts, domain, qtypes);
     }
 
     free(qtypes);
-    ldns_rdf_deep_free(domain);
+    if (domain != NULL)
+        ldns_rdf_deep_free(domain);
     ldns_resolver_deep_free(res);
 
     return EXIT_SUCCESS;
